@@ -19,7 +19,7 @@
  */
 
 #ifndef lint
-static  char rcsid[] = "@(#)$Id: ircd.c,v 1.62.2.8 2001/05/06 13:42:30 chopin Exp $";
+static const volatile char rcsid[] = "@(#)$Id: ircd.c,v 1.157 2005/03/29 22:50:15 chopin Exp $";
 #endif
 
 #include "os.h"
@@ -31,57 +31,49 @@ static  char rcsid[] = "@(#)$Id: ircd.c,v 1.62.2.8 2001/05/06 13:42:30 chopin Ex
 aClient me;			/* That's me */
 aClient *client = &me;		/* Pointer to beginning of Client list */
 
-static	void	open_debugfile(), setup_signals(), io_loop();
+static	void	open_debugfile(void), setup_signals(void), io_loop(void);
+
+#if defined(USE_IAUTH)
+static	RETSIGTYPE	s_slave(int s);
+#endif
 
 istat_t	istat;
+iconf_t iconf;
 char	**myargv;
 int	rehashed = 0;
 int	portnum = -1;		    /* Server port number, listening this */
 char	*configfile = IRCDCONF_PATH;	/* Server configuration file */
 int	debuglevel = -1;		/* Server debug level */
 int	bootopt = BOOT_PROT|BOOT_STRICTPROT;	/* Server boot option flags */
+int	serverbooting = 1;
+int	firstrejoindone = 0;		/* Server rejoined the network after
+					   start */
 char	*debugmode = "";		/*  -"-    -"-   -"-   -"- */
 char	*sbrk0;				/* initial sbrk(0) */
 char	*tunefile = IRCDTUNE_PATH;
-static	int	dorehash = 0,
-		dorestart = 0,
-		restart_iauth = 0;
+volatile static	int	dorehash = 0,
+			dorestart = 0,
+			restart_iauth = 0;
 
+#ifdef DELAY_CLOSE
+time_t	nextdelayclose = 0;	/* time for next delayed close */
+#endif
 time_t	nextconnect = 1;	/* time for next try_connections call */
 time_t	nextgarbage = 1;        /* time for next collect_channel_garbage call*/
 time_t	nextping = 1;		/* same as above for check_pings() */
 time_t	nextdnscheck = 0;	/* next time to poll dns to force timeouts */
 time_t	nextexpire = 1;		/* next expire run on the dns cache */
 time_t	nextiarestart = 1;	/* next time to check if iauth is alive */
+time_t	nextpreference = 1;	/* time for next calculate_preference call */
+#ifdef TKLINE
+time_t	nexttkexpire = 0;	/* time for next tkline_expire call */
+#endif
 
-#ifdef	PROFIL
-extern	etext();
+aClient *ListenerLL = NULL;	/* Listeners linked list */
 
-RETSIGTYPE	s_monitor(s)
-int s;
-{
-	static	int	mon = 0;
-#if POSIX_SIGNALS
-	struct	sigaction act;
-#endif
-
-	(void)moncontrol(mon);
-	mon = 1 - mon;
-#if POSIX_SIGNALS
-	act.sa_handler = s_rehash;
-	act.sa_flags = 0;
-	(void)sigemptyset(&act.sa_mask);
-	(void)sigaddset(&act.sa_mask, SIGUSR1);
-	(void)sigaction(SIGUSR1, &act, NULL);
-#else
-	(void)signal(SIGUSR1, s_monitor);
-#endif
-}
-#endif
-
-RETSIGTYPE s_die(s)
-int s;
+RETSIGTYPE s_die(int s)
 {
+	sendto_serv_v(NULL, SV_UID, ":%s SDIE", me.serv->sid);
 #ifdef	USE_SYSLOG
 	(void)syslog(LOG_CRIT, "Server Killed By SIGTERM");
 	(void)closelog();
@@ -93,8 +85,7 @@ int s;
 }
 
 #if defined(USE_IAUTH)
-RETSIGTYPE s_slave(s)
-int s;
+static	RETSIGTYPE	s_slave(int s)
 {
 # if POSIX_SIGNALS
 	struct	sigaction act;
@@ -111,8 +102,7 @@ int s;
 }
 #endif
 
-static RETSIGTYPE s_rehash(s)
-int s;
+static RETSIGTYPE s_rehash(int s)
 {
 #if POSIX_SIGNALS
 	struct	sigaction act;
@@ -128,8 +118,7 @@ int s;
 	dorehash = 1;
 }
 
-void	restart(mesg)
-char	*mesg;
+void	restart(char *mesg)
 {
 #ifdef	USE_SYSLOG
 	(void)syslog(LOG_WARNING, "Restarting Server because: %s (%u)", mesg,
@@ -140,8 +129,7 @@ char	*mesg;
 	server_reboot();
 }
 
-RETSIGTYPE s_restart(s)
-int s;
+RETSIGTYPE s_restart(int s)
 {
 #if POSIX_SIGNALS
 	struct	sigaction act;
@@ -152,12 +140,19 @@ int s;
 	(void)sigaddset(&act.sa_mask, SIGINT);
 	(void)sigaction(SIGINT, &act, NULL);
 #else
+
 	(void)signal(SIGHUP, SIG_DFL);	/* sysV -argv */
 #endif
+	if (bootopt & BOOT_TTY)
+	{
+		fprintf(stderr, "Caught SIGINT, terminating...\n");
+		exit(-1);
+	}
+
 	dorestart = 1;
 }
 
-void	server_reboot()
+void	server_reboot(void)
 {
 	Reg	int	i;
 
@@ -177,12 +172,14 @@ void	server_reboot()
 	for (i = 3; i < MAXCONNECTIONS; i++)
 		(void)close(i);
 	if (!(bootopt & (BOOT_TTY|BOOT_DEBUG)))
+	{
 		(void)close(2);
-	(void)close(1);
+		(void)close(1);
+	}
 	if ((bootopt & BOOT_CONSOLE) || isatty(0))
 		(void)close(0);
 	ircd_writetune(tunefile);
-	if (!(bootopt & (BOOT_INETD|BOOT_OPER)))
+	if (!(bootopt & BOOT_INETD))
 	    {
 		(void)execv(IRCD_PATH, myargv);
 #ifdef USE_SYSLOG
@@ -208,10 +205,8 @@ void	server_reboot()
 **	function should be made latest. (No harm done if this
 **	is called earlier or later...)
 */
-static	time_t	try_connections(currenttime)
-time_t	currenttime;
+static	time_t	try_connections(time_t currenttime)
 {
-	static	time_t	lastsort = 0;
 	Reg	aConfItem *aconf;
 	Reg	aClient *cptr;
 	aConfItem **pconf;
@@ -219,63 +214,108 @@ time_t	currenttime;
 	time_t	next = 0;
 	aClass	*cltmp;
 	aConfItem *con_conf = NULL;
-	double	f, f2;
-	aCPing	*cp;
+	int	allheld = 1;
+#ifdef DISABLE_DOUBLE_CONNECTS
+	int	i;
+#endif
 
 	Debug((DEBUG_NOTICE,"Connection check at   : %s",
 		myctime(currenttime)));
 	for (aconf = conf; aconf; aconf = aconf->next )
-	    {
-		/* Also when already connecting! (update holdtimes) --SRB */
+	{
+		/* not a C-line */
 		if (!(aconf->status & (CONF_CONNECT_SERVER|CONF_ZCONNECT_SERVER)))
 			continue;
-		/*
-		** Skip this entry if the use of it is still on hold until
-		** future. Otherwise handle this entry (and set it on hold
-		** until next time). Will reset only hold times, if already
-		** made one successfull connection... [this algorithm is
-		** a bit fuzzy... -- msa >;) ]
-		*/
-		if ((aconf->hold > currenttime))
-		    {
-			if ((next > aconf->hold) || (next == 0))
-				next = aconf->hold;
-			continue;
-		    }
-		send_ping(aconf);
+
+		/* not a candidate for AC */
 		if (aconf->port <= 0)
 			continue;
 
+		/* minimize next to lowest hold time of all AC-able C-lines */
+		if (next > aconf->hold || next == 0)
+			next = aconf->hold;
+
+		/* skip conf if the use of it is on hold until future. */
+		if (aconf->hold > currenttime)
+			continue;
+
+		/* at least one candidate not held for future, good */
+		allheld = 0;
+
 		cltmp = Class(aconf);
+		/* see if another link in this conf is allowed */
+		if (Links(cltmp) >= MaxLinks(cltmp))
+			continue;
+		
+		/* next possible check after connfreq secs for this C-line */
 		confrq = get_con_freq(cltmp);
 		aconf->hold = currenttime + confrq;
-		/*
-		** Found a CONNECT config with port specified, scan clients
-		** and see if this server is already connected?
-		*/
+
+		/* is this server already connected? */
 		cptr = find_name(aconf->name, (aClient *)NULL);
 		if (!cptr)
 			cptr = find_mask(aconf->name, (aClient *)NULL);
-		/*
-		** It is not connected, scan clients and see if any matches
-		** a D(eny) line.
-		*/
+
+		/* matching client already exists, no AC to it */
+		if (cptr)
+			continue;
+
+		/* no such server, check D-lines */
 		if (find_denied(aconf->name, Class(cltmp)))
 			continue;
-		/* We have a candidate, let's see if it could be the best. */
-		if (!cptr && (Links(cltmp) < MaxLinks(cltmp)) &&
-		    (!con_conf ||
+
+#ifdef DISABLE_DOUBLE_CONNECTS
+		/* Much better would be traversing only unknown
+		** connections, but this requires another global
+		** variable, adding and removing from there in
+		** proper places etc. Some day. --B. */
+		for (i = highest_fd; i >= 0; i--)
+		{
+			if (!(cptr = local[i]) ||
+				cptr->status > STAT_UNKNOWN)
+			{
+				continue;
+			}
+			/* an unknown traveller we have */
+			if (
+#ifndef INET6
+				cptr->ip.s_addr == aconf->ipnum.s_addr
+#else
+				!memcmp(cptr->ip.s6_addr,
+					aconf->ipnum.s6_addr, 16)
+#endif
+			)
+			{
+				/* IP the same. Coincidence? Maybe.
+				** Do not cause havoc with double connect. */
+				break;
+			}
+			cptr = NULL;
+		}
+		if (cptr)
+		{
+			sendto_flag(SCH_SERVER, "AC to %s postponed", aconf->name);
+			continue;
+		}
+#endif
+		/* we have a candidate! */
+
+		/* choose the best. */
+		if (!con_conf ||
 		     (con_conf->pref > aconf->pref && aconf->pref >= 0) ||
 		     (con_conf->pref == -1 &&
-		      Class(cltmp) > ConfClass(con_conf))))
+		      Class(cltmp) > ConfClass(con_conf)))
+		{
 			con_conf = aconf;
-		if ((next > aconf->hold) || (next == 0))
-			next = aconf->hold;
-	    }
+		}
+		/* above is my doubt: if we always choose best connection
+		** and it always fails connecting, we may never try another,
+		** even "worse"; what shall we do? --Beeth */
+	}
 	if (con_conf)
-	    {
+	{
 		if (con_conf->next)  /* are we already last? */
-		    {
+		{
 			for (pconf = &conf; (aconf = *pconf);
 			     pconf = &(aconf->next))
 				/* put the current one at the end and
@@ -284,110 +324,203 @@ time_t	currenttime;
 				if (aconf == con_conf)
 					*pconf = aconf->next;
 			(*pconf = con_conf)->next = 0;
-		    }
-		if (connect_server(con_conf, (aClient *)NULL,
+		}
+
+		/* "Penalty" for being the best, so in next call of
+		 * try_connections() other servers have chance. --B. */
+		con_conf->hold += get_con_freq(Class(con_conf));
+
+		if (!iconf.aconnect)
+		{
+			sendto_flag(SCH_NOTICE,
+				"Connection to %s deferred. Autoconnect "
+				"administratively disabled", con_conf->name);
+		}
+		else if (connect_server(con_conf, (aClient *)NULL,
 				   (struct hostent *)NULL) == 0)
+		{
 			sendto_flag(SCH_NOTICE,
 				    "Connection to %s[%s] activated.",
 				    con_conf->name, con_conf->host);
-	    }
+		}
+	}
+	else
+	if (allheld == 0)	/* disable AC only when some C: got checked */
+	{
+		/* No suitable conf for AC was found, so why bother checking
+		** again? If some server quits, it'd get reenabled --B. */
+		next = 0;
+	}
 	Debug((DEBUG_NOTICE,"Next connection check : %s", myctime(next)));
-	/*
-	 * calculate preference value based on accumulated stats.
-	 */
-	if (!lastsort || lastsort < currenttime)
-	    {
-		for (aconf = conf; aconf; aconf = aconf->next)
-			if (!(cp = aconf->ping) || !cp->seq || !cp->recvd)
-				aconf->pref = -1;
-			else
-			    {
-				f = (double)cp->recvd / (double)cp->seq;
-				f2 = pow(f, (double)20.0);
-				if (f2 < (double)0.001)
-					f = (double)0.001;
-				else
-					f = f2;
-				f2 = (double)cp->ping / (double)cp->recvd;
-				f = f2 / f;
-				if (f > 100000.0)
-					f = 100000.0;
-				aconf->pref = (u_int) (f * (double)100.0);
-			    }
-		lastsort = currenttime + 60;
-	    }
 	return (next);
 }
 
-
-static	void	close_held(cptr)
-aClient	*cptr;
+/*
+ * calculate preference value based on accumulated stats.
+ */
+time_t calculate_preference(time_t currenttime)
 {
-	Reg	aClient	*acptr;
-	int	i;
+	aConfItem *aconf;
+	aCPing	*cp;
+	double	f, f2;
 
-	for (i = highest_fd; i >= 0; i--)
-		if ((acptr = local[i]) && (cptr->port == acptr->port) &&
-		    (acptr != cptr) && IsHeld(acptr) &&
-		    !bcmp((char *)&cptr->ip, (char *)&acptr->ip,
-			  sizeof(acptr->ip)))
-		    {
-			(void) exit_client(acptr, acptr, &me,
-					   "Reconnect Timeout");
-			return;
-		    }
+	for (aconf = conf; aconf; aconf = aconf->next)
+	{
+		/* not a C-line */
+		if (!(aconf->status & (CONF_CONNECT_SERVER|CONF_ZCONNECT_SERVER)))
+			continue;
+
+		/* not a candidate for AC */
+		if (aconf->port <= 0)
+			continue;
+
+		/* send (udp) pings for all AC-able C-lines, we'll use it to
+		** calculate preferences */
+		send_ping(aconf);
+
+		if (!(cp = aconf->ping) || !cp->seq || !cp->recvd)
+		{
+			aconf->pref = -1;
+		}
+		else
+		{
+			f = (double)cp->recvd / (double)cp->seq;
+			f2 = pow(f, (double)20.0);
+			if (f2 < (double)0.001)
+				f = (double)0.001;
+			else
+				f = f2;
+			f2 = (double)cp->ping / (double)cp->recvd;
+			f = f2 / f;
+			if (f > 100000.0)
+				f = 100000.0;
+			aconf->pref = (u_int) (f * (double)100.0);
+		}
+	}
+	return currenttime + 60;
 }
 
-
-static	time_t	check_pings(currenttime)
-time_t	currenttime;
+/* Checks all clients against KILL lines. (And remove them, if found.)
+** Only MAXDELAYEDKILLS at a time or all, if not defined.
+** Returns 1, if still work to do, 0 if finished.
+*/
+static	int	delayed_kills(time_t currenttime)
 {
+	static	time_t	dk_rehashed = 0;	/* time of last rehash we're processing */
+	static	int	dk_lastfd;		/* fd we last checked */
+	static	int	dk_checked;		/* # clients we checked */
+	static	int	dk_killed;		/* # clients we killed */
+	Reg	aClient	*cptr;
+	Reg	int	i, j;
+
+	if (dk_rehashed == 0)
+	{
+		dk_rehashed = currenttime;
+		dk_checked = 0;
+		dk_killed = 0;
+		dk_lastfd = highest_fd;
+	}
+#ifdef MAXDELAYEDKILLS
+	/* checking only this many clients each time */
+	j = dk_lastfd - MAXDELAYEDKILLS + 1;
+	if (j < 0)
+#endif
+		j = 0;
+
+	for (i = dk_lastfd; i >= j; i--)
+	{
+		int	kflag = 0;
+		char	*reason = NULL;
+
+		if (!(cptr = local[i]) || !IsPerson(cptr))
+		{
+			/* for K:lines we're interested only in local,
+			** fully registered clients */
+			if (j > 0)
+				j--;
+			continue;
+		}
+
+		dk_checked++;
+		kflag = find_kill(cptr, 0, &reason);
+
+		/* If the client is a user and a KILL line was found
+		** to be active, close this connection. */
+		if (kflag == -1)
+		{
+			char buf[100];
+
+			dk_killed++;
+			sendto_flag(SCH_NOTICE,
+				"Kill line active for %s",
+				get_client_name(cptr, FALSE));
+				cptr->exitc = EXITC_KLINE;
+			if (!BadPtr(reason))
+				sprintf(buf, "Kill line active: %.80s",
+						reason);
+			(void)exit_client(cptr, cptr, &me, (reason) ?
+					  buf : "Kill line active");
+		}
+	}
+	dk_lastfd = i;	/* from which fd to start next time */
+	Debug((DEBUG_DEBUG, "DelayedKills killed %d and counting...",
+		dk_killed));
+
+	if (dk_lastfd < 0)
+	{
+		sendto_flag(SCH_NOTICE, "DelayedKills checked %d killed %d "
+			"in %d sec", dk_checked, dk_killed,
+			currenttime - dk_rehashed);
+		dk_rehashed = 0;
+		if (rehashed == 2)
+		{
+			/* there was rehash queued, start again */
+			return 1;
+		}
+		return 0;
+	}
+	return rehashed;
+}
+
+static	time_t	check_pings(time_t currenttime)
+{
+#ifdef TIMEDKLINES
 	static	time_t	lkill = 0;
+#endif
 	Reg	aClient	*cptr;
 	Reg	int	kflag = 0;
-	int	ping = 0, i, rflag = 0;
+	aClient *bysptr = NULL;
+	int	ping = 0, i;
 	time_t	oldest = 0, timeout;
-	char	*reason;
+	char	*reason = NULL;
 
 	for (i = highest_fd; i >= 0; i--)
 	    {
-		if (!(cptr = local[i]) || IsListening(cptr) || IsLog(cptr) ||
-		    IsHeld(cptr))
+		if (!(cptr = local[i]) || IsListener(cptr))
 			continue;
 
-		/*
-		 * K and R lines once per minute, max.  This is the max.
-		 * granularity in K-lines anyway (with time field).
-		 */
-		if (
-#if defined(TIMEDKLINES) || ( defined(R_LINES) && defined(R_LINES_OFTEN) )
-			(currenttime - lkill > TIMEDKLINES) || 
-#endif /* TIMEDKLINES */
-			rehashed)
-		    {
-			if (IsPerson(cptr))
-			    {
-				kflag = find_kill(cptr, rehashed, &reason);
-#ifdef R_LINES_OFTEN
-				rflag = find_restrict(cptr);
+#ifdef TIMEDKLINES
+		kflag = 0;
+		reason = NULL;
+		/* 
+		** Once per TIMEDKLINES seconds.
+		** (1 minute is minimum resolution in K-line field)
+		*/
+		if ((currenttime - lkill > TIMEDKLINES)
+			&& IsPerson(cptr) && !IsKlineExempt(cptr))
+		{
+			kflag = find_kill(cptr, 1, &reason);
+		}
 #endif
-			    }
-			else
-			    {
-				kflag = rflag = 0;
-				reason = NULL;
-			    }
-		    }
-		ping = IsRegistered(cptr) ? get_client_ping(cptr) :
-					    ACCEPTTIMEOUT;
-		Debug((DEBUG_DEBUG, "c(%s) %d p %d k %d r %d a %d",
-			cptr->name, cptr->status, ping, kflag, rflag,
+		ping = IsRegistered(cptr) ? cptr->ping : ACCEPTTIMEOUT;
+		Debug((DEBUG_DEBUG, "c(%s) %d p %d k %d a %d",
+			cptr->name, cptr->status, ping, kflag,
 			currenttime - cptr->lasttime));
 		/*
 		 * Ok, so goto's are ugly and can be avoided here but this code
 		 * is already indented enough so I think its justified. -avalon
 		 */
-		if (!kflag && !rflag && IsRegistered(cptr) &&
+		if (!kflag && IsRegistered(cptr) &&
 		    (ping >= currenttime - cptr->lasttime))
 			goto ping_timeout;
 		/*
@@ -396,21 +529,12 @@ time_t	currenttime;
 		 * If the client is a user and a KILL line was found
 		 * to be active, close this connection too.
 		 */
-		if (kflag || rflag ||
+		if (kflag ||
 		    ((currenttime - cptr->lasttime) >= (2 * ping) &&
 		     (cptr->flags & FLAGS_PINGSENT)) ||
 		    (!IsRegistered(cptr) &&
 		     (currenttime - cptr->firsttime) >= ping))
 		    {
-			if (IsReconnect(cptr))
-			    {
-				sendto_flag(SCH_ERROR,
-					    "Reconnect timeout to %s",
-					    get_client_name(cptr, TRUE));
-				close_held(cptr);
-				(void)exit_client(cptr, cptr, &me,
-						  "Ping timeout");
-			    }
 			if (!IsRegistered(cptr) && 
 			    (DoingDNS(cptr) || DoingAuth(cptr) ||
 			     DoingXAuth(cptr)))
@@ -446,7 +570,7 @@ time_t	currenttime;
 					    {
 						cptr->exitc = EXITC_AUTHTOUT;
 						sendto_iauth("%d T", cptr->fd);
-						ereject_user(cptr, " Timeout ",
+						exit_client(cptr, cptr, &me,
 						     "Authentication Timeout");
 						continue;
 					    }
@@ -463,12 +587,27 @@ time_t	currenttime;
 			    }
 			if (IsServer(cptr) || IsConnecting(cptr) ||
 			    IsHandshake(cptr))
+			{
+				if (cptr->serv && cptr->serv->byuid[0])
+				{
+					bysptr = find_uid(cptr->serv->byuid,
+							NULL);
+				}
+				/* we are interested only in *remote* opers */
+				if (bysptr && !MyConnect(bysptr))
+				{
+					sendto_one(bysptr, ":%s NOTICE %s :"
+						"No response from %s, closing"
+						" link", ME, bysptr->name,
+						get_client_name(cptr, FALSE));
+				}
 				sendto_flag(SCH_NOTICE,
 					    "No response from %s closing link",
 					    get_client_name(cptr, FALSE));
+			}
 			/*
 			 * this is used for KILL lines with time restrictions
-			 * on them - send a messgae to the user being killed
+			 * on them - send a message to the user being killed
 			 * first.
 			 */
 			if (kflag && IsPerson(cptr))
@@ -485,18 +624,6 @@ time_t	currenttime;
 				(void)exit_client(cptr, cptr, &me, (reason) ?
 						  buf : "Kill line active");
 			    }
-
-#if defined(R_LINES) && defined(R_LINES_OFTEN)
-			else if (IsPerson(cptr) && rflag)
-			    {
-				sendto_flag(SCH_NOTICE,
-					   "Restricting %s, closing link.",
-					   get_client_name(cptr,FALSE));
-				cptr->exitc = EXITC_RLINE;
-				(void)exit_client(cptr, cptr, &me,
-						  "Restricting");
-			    }
-#endif
 			else
 			    {
 				cptr->exitc = EXITC_PING;
@@ -525,20 +652,21 @@ ping_timeout:
 		if (timeout < oldest || !oldest)
 			oldest = timeout;
 	    }
+#ifdef TIMEDKLINES
 	if (currenttime - lkill > 60)
 		lkill = currenttime;
+#endif
 	if (!oldest || oldest < currenttime)
 		oldest = currenttime + PINGFREQUENCY;
-	if (oldest < currenttime + 2)
-		oldest += 2;
+	if (oldest < currenttime + 30)
+		oldest += 30;
 	Debug((DEBUG_NOTICE,"Next check_ping() call at: %s, %d %d %d",
 		myctime(oldest), ping, oldest, currenttime));
 	return (oldest);
 }
 
 
-static	void	setup_me(mp)
-aClient	*mp;
+static	void	setup_me(aClient *mp)
 {
 	struct	passwd	*p;
 
@@ -546,31 +674,10 @@ aClient	*mp;
 	strncpyzt(mp->username, (p) ? p->pw_name : "unknown",
 		  sizeof(mp->username));
 	(void)get_my_name(mp, mp->sockhost, sizeof(mp->sockhost)-1);
-
-	/* Setup hostp - fake record to resolve localhost. -Toor */
-	mp->hostp = (struct hostent *)MyMalloc(sizeof(struct hostent));
-	mp->hostp->h_name = MyMalloc(strlen(me.sockhost)+1);
-	strcpy(mp->hostp->h_name, mp->sockhost);
-	mp->hostp->h_aliases = (char **)MyMalloc(sizeof(char *));
-	*mp->hostp->h_aliases = NULL;
-	mp->hostp->h_addrtype = AFINET;
-	mp->hostp->h_length = 
-#ifdef	INET6
-				IN6ADDRSZ;
-#else
-				sizeof(long);
-#endif
-	mp->hostp->h_addr_list = (char **)MyMalloc(2*sizeof(char *));
-#ifdef	INET6
-	mp->hostp->h_addr_list[0] = (char *)&in6addr_loopback;
-#else
-	mp->hostp->h_addr_list[0] = (void *)MyMalloc(mp->hostp->h_length);
-	*(long *)(mp->hostp->h_addr_list[0]) = IN_LOOPBACKNET;
-#endif
-	mp->hostp->h_addr_list[1] = NULL ;
-
-	if (mp->name[0] == '\0')
-		strncpyzt(mp->name, mp->sockhost, sizeof(mp->name));
+	/* I think we need no hostp, especially fake one --B.  */
+	mp->hostp = NULL;
+	if (mp->serv->namebuf[0] == '\0')
+		strncpyzt(mp->serv->namebuf, mp->sockhost, sizeof(mp->serv->namebuf));
 	if (me.info == DefInfo)
 		me.info = mystrdup("IRCers United");
 	mp->lasttime = mp->since = mp->firsttime = time(NULL);
@@ -584,19 +691,24 @@ aClient	*mp;
 	mp->user = NULL;
 	mp->fd = -1;
 	SetMe(mp);
-	(void) make_server(mp);
 	mp->serv->snum = find_server_num (ME);
-	(void) make_user(mp);
+	/* we don't fill our own IP -> 0 as ip lenght */
+	(void) make_user(mp,0);
 	istat.is_users++;	/* here, cptr->next is NULL, see make_user() */
 	mp->user->flags |= FLAGS_OPER;
-	mp->serv->up = mp->name;
+	mp->serv->up = mp;
+	mp->serv->maskedby = mp;
+	mp->serv->version |= SV_UID;
 	mp->user->server = find_server_string(mp->serv->snum);
 	strncpyzt(mp->user->username, (p) ? p->pw_name : "unknown",
 		  sizeof(mp->user->username));
 	(void) strcpy(mp->user->host, mp->name);
+	SetEOB(mp);
+	istat.is_eobservers = 1;
 
 	(void)add_to_client_hash_table(mp->name, mp);
-
+	(void)add_to_sid_hash_table(mp->serv->sid, mp);
+	strncpyzt(mp->serv->verstr, PATCHLEVEL, sizeof(mp->serv->verstr));
 	setup_server_channels(mp);
 }
 
@@ -605,17 +717,18 @@ aClient	*mp;
 **	This is called when the commandline is not acceptable.
 **	Give error message and exit without starting anything.
 */
-static	int	bad_command()
+static	void	bad_command(void)
 {
   (void)printf(
-	 "Usage: ircd [-a] [-b] [-c]%s [-h servername] [-q] [-o] [-i] [-T tunefile] [-p (strict|on|off)] [-s] [-v] %s\n",
+	 "Usage: ircd [-a] [-b] [-c]%s [-h servername] [-q] [-i]"
+	 "[-T [tunefile]] [-p (strict|on|off)] [-s] [-v] [-t] %s\n",
 #ifdef CMDLINE_CONFIG
 	 " [-f config]",
 #else
 	 "",
 #endif
 #ifdef DEBUGMODE
-	 " [-x loglevel] [-t]"
+	 " [-x loglevel]"
 #else
 	 ""
 #endif
@@ -624,21 +737,13 @@ static	int	bad_command()
   exit(-1);
 }
 
-int	main(argc, argv)
-int	argc;
-char	*argv[];
+int	main(int argc, char *argv[])
 {
 	uid_t	uid, euid;
 
-	(void) myctime(time(NULL));	/* Don't ask, just *don't* ask */
 	sbrk0 = (char *)sbrk((size_t)0);
 	uid = getuid();
 	euid = geteuid();
-#ifdef	PROFIL
-	(void)monstartup(0, etext);
-	(void)moncontrol(1);
-	(void)signal(SIGUSR1, s_monitor);
-#endif
 
 #ifdef	CHROOTDIR
 	ircd_res_init();
@@ -679,6 +784,9 @@ char	*argv[];
 	(void)umask(077);                /* better safe than sorry --SRB */
 	bzero((char *)&me, sizeof(me));
 
+	make_server(&me);
+	register_server(&me);
+
 	version = make_version();	/* Generate readable version string */
 
 	/*
@@ -693,13 +801,17 @@ char	*argv[];
 		int	flag = *p++;
 
 		if (flag == '\0' || *p == '\0')
+		{
 			if (argc > 1 && argv[1][0] != '-')
-			    {
+			{
 				p = *++argv;
 				argc -= 1;
-			    }
+			}
 			else
+			{
 				p = "";
+			}
+		}
 
 		switch (flag)
 		    {
@@ -715,10 +827,6 @@ char	*argv[];
 		    case 'q':
 			bootopt |= BOOT_QUICK;
 			break;
-		    case 'o': /* Per user local daemon... */
-                        (void)setuid((uid_t)uid);
-			bootopt |= BOOT_OPER;
-		        break;
 #ifdef CMDLINE_CONFIG
 		    case 'f':
                         (void)setuid((uid_t)uid);
@@ -728,7 +836,7 @@ char	*argv[];
 		    case 'h':
 			if (*p == '\0')
 				bad_command();
-			strncpyzt(me.name, p, sizeof(me.name));
+			strncpyzt(me.serv->namebuf, p, sizeof(me.serv->namebuf));
 			break;
 		    case 'i':
 			bootopt |= BOOT_INETD|BOOT_AUTODIE;
@@ -747,12 +855,12 @@ char	*argv[];
 			bootopt |= BOOT_NOIAUTH;
 			break;
 		    case 't':
+#ifdef DEBUGMODE
                         (void)setuid((uid_t)uid);
+#endif
 			bootopt |= BOOT_TTY;
 			break;
 		    case 'T':
-			if (*p == '\0')
-				bad_command();
 			tunefile = p;
 			break;
 		    case 'v':
@@ -783,17 +891,14 @@ char	*argv[];
 		    }
 	    }
 
+	if (strlen(tunefile) > 1023 || strlen(mybasename(tunefile)) > 42)
+	{
+		fprintf(stderr, "Too long tune filename\n");
+		exit(-1);
+	}
 	if (argc > 0)
 		bad_command(); /* This exits out */
 
-#if defined(USE_IAUTH) && defined(__CYGWIN32__)
-	if ((bootopt & BOOT_NOIAUTH) == 0)
-	    {
-		bootopt |= BOOT_NOIAUTH;
-		(void)fprintf(stderr, "WARNING: Assuming -s option.\n");
-	    }
-#endif
-	
 #ifndef IRC_UID
 	if ((uid != euid) && !euid)
 	    {
@@ -820,6 +925,8 @@ char	*argv[];
 #endif /*CHROOTDIR/UID/GID*/
 
 #if defined(USE_IAUTH)
+	/* At this point, we just check whether iauth is there. Real start
+	 * is done in init_sys(). */
 	if ((bootopt & BOOT_NOIAUTH) == 0)
 		switch (vfork())
 		    {
@@ -850,6 +957,8 @@ char	*argv[];
 
 	/* didn't set debuglevel */
 	/* but asked for debugging output to tty */
+#ifdef DEBUGMODE
+
 	if ((debuglevel < 0) &&  (bootopt & BOOT_TTY))
 	    {
 		(void)fprintf(stderr,
@@ -857,13 +966,14 @@ char	*argv[];
 		exit(-1);
 	    }
 
-	initstats();
-	ircd_readtune(tunefile);
+#endif
 	timeofday = time(NULL);
-#ifdef	CACHED_MOTD
+	initanonymous();
+	initstats();
+	initruntimeconf();
+	ircd_readtune(tunefile);
 	motd = NULL;
 	read_motd(IRCDMOTD_PATH);
-#endif
 	inithashtables();
 	initlists();
 	initclass();
@@ -872,7 +982,6 @@ char	*argv[];
 	open_debugfile();
 	timeofday = time(NULL);
 	(void)init_sys();
-	logfiles_open();
 
 #ifdef USE_SYSLOG
 	openlog(mybasename(myargv[0]), LOG_PID|LOG_NDELAY, LOG_FACILITY);
@@ -880,12 +989,12 @@ char	*argv[];
 	timeofday = time(NULL);
 	if (initconf(bootopt) == -1)
 	    {
-		Debug((DEBUG_FATAL, "Failed in reading configuration file %s",
+		Debug((DEBUG_FATAL, "Couldn't open configuration file %s",
 			configfile));
-		/* no can do.
-		(void)printf("Couldn't open configuration file %s\n",
-			configfile);
-		*/
+		(void)fprintf(stderr,
+			"Couldn't open configuration file %s (%s)\n",
+			 configfile,strerror(errno));
+		
 		exit(-1);
 	    }
 	else
@@ -895,21 +1004,53 @@ char	*argv[];
 
                 for (i = 0; i <= highest_fd; i++)
                     {   
-                        if (!(acptr = local[i]))
-                                continue;
-			if (IsListening(acptr))
+			if (!(acptr = local[i]))
+				continue;
+			if (IsListener(acptr))
 				break;
 			acptr = NULL;
 		    }
 		/* exit if there is nothing to listen to */
 		if (acptr == NULL && !(bootopt & BOOT_INETD))
+		{
+			fprintf(stderr,
+			"Fatal Error: No working P-line in ircd.conf\n");
 			exit(-1);
+		}
 		/* Is there an M-line? */
 		if (!find_me())
+		{
+			fprintf(stderr,
+			"Fatal Error: No M-line in ircd.conf.\n");
 			exit(-1);
+		}
+		if (check_servername(ME))
+		{
+			fprintf(stderr,
+			"Fatal Error: Invalid server name.\n");
+			exit(-1);
+		}
+		if (!me.serv->sid)
+		{
+			fprintf(stderr,
+			"Fatal Error: No SID specified in ircd.conf\n");
+			exit(-1);
+		}
+		if (!sid_valid(me.serv->sid))
+		{
+			fprintf(stderr,
+			"Fatal Error: Invalid sid %s specified in ircd.conf\n",
+				me.serv->sid);
+			exit(-1);
+		}
+		if (!networkname)
+		{
+			fprintf(stderr,
+			"Warning: Network name is not set in ircd.conf\n");
+		}
+		isupport = make_isupport();	/* Generate RPL_ISUPPORT (005) numerics */
 	    }
 
-	dbuf_init();
 	setup_me(&me);
 	check_class();
 	ircd_writetune(tunefile);
@@ -919,6 +1060,8 @@ char	*argv[];
 		aConfItem *aconf;
 
 		tmp = make_client(NULL);
+		make_server(tmp);
+		register_server(tmp);
 
 		tmp->fd = 0;
 		tmp->flags = FLAGS_LISTEN;
@@ -928,9 +1071,9 @@ char	*argv[];
 
                 SetMe(tmp);
 
-                (void)strcpy(tmp->name, "*");
+                (void)strcpy(tmp->serv->namebuf, "*");
 
-                if (inetport(tmp, 0, "0.0.0.0", 0))
+                if (inetport(tmp, 0, "0.0.0.0", 0, 1))
                         tmp->fd = -1;
 		if (tmp->fd == 0) 
 		    {
@@ -950,34 +1093,44 @@ char	*argv[];
 		else
 		    exit(5);
 	    }
-	if (bootopt & BOOT_OPER)
-	    {
-		aClient *tmp = add_connection(&me, 0);
-
-		if (!tmp)
-			exit(1);
-		SetMaster(tmp);
-		local[0] = tmp;
-	    }
-	else
-		write_pidfile();
-
+	
 	Debug((DEBUG_NOTICE,"Server ready..."));
 #ifdef USE_SYSLOG
 	syslog(LOG_NOTICE, "Server Ready: v%s (%s #%s)", version, creation,
 	       generation);
 #endif
+	printf("Server %s (%s) version %s starting%s%s", ME, me.serv->sid,
+		version, (bootopt & BOOT_TTY) ? " in foreground mode." : ".",
+#ifdef DEBUGMODE
+		"(DEBUGMODE)\n"
+#else
+		"\n"
+#endif
+		);
+
 	timeofday = time(NULL);
+	mysrand(timeofday);
+	
+	/* daemonize() closes 0,1,2 -- make sure you don't have any fd open */
+	daemonize();	
+	logfiles_open();
+	write_pidfile();
+	dbuf_init();
+	
+	serverbooting = 0;
+	
 	while (1)
 		io_loop();
 }
 
 
-void	io_loop()
+static	void	io_loop(void)
 {
 	static	time_t	delay = 0;
 	int maxs = 4;
 
+	if (timeofday >= nextpreference)
+		nextpreference = calculate_preference(timeofday);
 	/*
 	** We only want to connect if a connection is due,
 	** not every time through.  Note, if there are no
@@ -986,9 +1139,19 @@ void	io_loop()
 	*/
 	if (nextconnect && timeofday >= nextconnect)
 		nextconnect = try_connections(timeofday);
+#ifdef DELAY_CLOSE
+	/* close all overdue delayed fds */
+	if (nextdelayclose && timeofday >= nextdelayclose)
+		nextdelayclose = delay_close(-1);
+#endif
+#ifdef TKLINE
+	/* expire tklines */
+	if (nexttkexpire && timeofday >= nexttkexpire)
+		nexttkexpire = tkline_expire(0);
+#endif
 	/*
 	** Every once in a while, hunt channel structures that
-	** can be freed.
+	** can be freed. Reop channels while at it, too.
 	*/
 	if (timeofday >= nextgarbage)
 		nextgarbage = collect_channel_garbage(timeofday);
@@ -1008,8 +1171,13 @@ void	io_loop()
 		delay = MIN(nextping, nextconnect);
 	else
 		delay = nextping;
+#ifdef DELAY_CLOSE
+	if (nextdelayclose)
+		delay = MIN(nextdelayclose, delay);
+#endif
 	delay = MIN(nextdnscheck, delay);
 	delay = MIN(nextexpire, delay);
+	delay = MIN(nextpreference, delay);
 	delay -= timeofday;
 	/*
 	** Adjust delay to something reasonable [ad hoc values]
@@ -1026,9 +1194,8 @@ void	io_loop()
 		delay = MIN(delay, TIMESEC);
 
 	/*
-	** First, try to drain traffic from servers (this includes listening
-	** ports).  Give up, either if there's no traffic, or too many
-	** iterations.
+	** First, try to drain traffic from servers and listening sockets.
+	** Give up either if there's no traffic or too many iterations.
 	*/
 	while (maxs--)
 		if (read_message(0, &fdas, 0))
@@ -1063,7 +1230,10 @@ void	io_loop()
 	if (timeofday >= nextping)
 	    {
 		nextping = check_pings(timeofday);
-		rehashed = 0;
+		if (rehashed > 0)
+		{
+			rehashed = delayed_kills(timeofday);
+		}
 	    }
 
 	if (dorestart)
@@ -1103,23 +1273,13 @@ void	io_loop()
  * set from the command line by -x, use /dev/null as the dummy logfile as long
  * as DEBUGMODE has been defined, else don't waste the fd.
  */
-static	void	open_debugfile()
+static	void	open_debugfile(void)
 {
 #ifdef	DEBUGMODE
 	int	fd;
-	aClient	*cptr;
 
 	if (debuglevel >= 0)
 	    {
-		cptr = make_client(NULL);
-		cptr->fd = 2;
-		SetLog(cptr);
-		cptr->port = debuglevel;
-		cptr->flags = 0;
-		cptr->acpt = cptr;
-		local[2] = cptr;
-		(void)strcpy(cptr->sockhost, me.sockhost);
-
 		(void)printf("isatty = %d ttyname = %#x\n",
 			isatty(2), (u_int)ttyname(2));
 		if (!(bootopt & BOOT_TTY)) /* leave debugging output on fd 2 */
@@ -1133,22 +1293,17 @@ static	void	open_debugfile()
 				(void)dup2(fd, 2);
 				(void)close(fd); 
 			    }
-			strncpyzt(cptr->name, IRCDDBG_PATH,sizeof(cptr->name));
 		    }
-		else if (isatty(2) && ttyname(2))
-			strncpyzt(cptr->name, ttyname(2), sizeof(cptr->name));
-		else
-			(void)strcpy(cptr->name, "FD2-Pipe");
 		Debug((DEBUG_FATAL, "Debug: File <%s> Level: %d at %s",
-			cptr->name, cptr->port, myctime(time(NULL))));
+			( (!(bootopt & BOOT_TTY)) ? IRCDDBG_PATH :
+			(isatty(2) && ttyname(2)) ? ttyname(2) : "FD2-Pipe"),
+			debuglevel, myctime(time(NULL))));
 	    }
-	else
-		local[2] = NULL;
 #endif
 	return;
 }
 
-static	void	setup_signals()
+static	void	setup_signals(void)
 {
 #if POSIX_SIGNALS
 	struct	sigaction act;
@@ -1177,8 +1332,12 @@ static	void	setup_signals()
 	(void)sigaction(SIGTERM, &act, NULL);
 # if defined(USE_IAUTH)
 	act.sa_handler = s_slave;
+# else
+	act.sa_handler = SIG_IGN;
+# endif
 	(void)sigaddset(&act.sa_mask, SIGUSR1);
 	(void)sigaction(SIGUSR1, &act, NULL);
+# if defined(USE_IAUTH)
 	act.sa_handler = SIG_IGN;
 #  ifdef SA_NOCLDWAIT
         act.sa_flags = SA_NOCLDWAIT;
@@ -1188,6 +1347,15 @@ static	void	setup_signals()
 	(void)sigaddset(&act.sa_mask, SIGCHLD);
 	(void)sigaction(SIGCHLD, &act, NULL);
 # endif
+	
+# if defined(__FreeBSD__)	
+	/* Don't core after detaching from gdb on fbsd */
+
+	act.sa_handler = SIG_IGN;
+	act.sa_flags = 0;
+	(void)sigaddset(&act.sa_mask, SIGTRAP);
+	(void)sigaction(SIGTRAP,&act,NULL);
+# endif /* __FreeBSD__ */
 
 #else /* POSIX_SIGNALS */
 
@@ -1201,15 +1369,24 @@ static	void	setup_signals()
 	(void)signal(SIGWINCH, SIG_IGN);
 #  endif
 	(void)signal(SIGPIPE, SIG_IGN);
+	
 # endif /* HAVE_RELIABLE_SIGNALS */
-	(void)signal(SIGALRM, dummy);   
+	(void)signal(SIGALRM, dummy);
 	(void)signal(SIGHUP, s_rehash);
-	(void)signal(SIGTERM, s_die); 
+	(void)signal(SIGTERM, s_die);
 	(void)signal(SIGINT, s_restart);
 # if defined(USE_IAUTH)
 	(void)signal(SIGUSR1, s_slave);
 	(void)signal(SIGCHLD, SIG_IGN);
+# else
+	(void)signal(SIGUSR1, SIG_IGN);
 # endif
+	
+# if defined(__FreeBSD__)
+	/* don't core after detaching from gdb on fbsd */
+	(void)signal(SIGTRAP, SIG_IGN);
+# endif /* __FreeBSD__ */
+
 #endif /* POSIX_SIGNAL */
 
 #ifdef RESTARTING_SYSTEMCALLS
@@ -1226,18 +1403,20 @@ static	void	setup_signals()
  * Called from bigger_hash_table(), s_die(), server_reboot(),
  * main(after initializations), grow_history(), rehash(io_loop) signal.
  */
-void ircd_writetune(filename)
-char *filename;
+void ircd_writetune(char *filename)
 {
 	int fd;
 	char buf[100];
 
+	if (!filename || !*filename)
+		return;
+
 	(void)truncate(filename, 0);
 	if ((fd = open(filename, O_CREAT|O_WRONLY, 0600)) >= 0)
 	    {
-		(void)sprintf(buf, "%d\n%d\n%d\n%d\n%d\n%d\n", ww_size,
+		(void)sprintf(buf, "%d\n%d\n%d\n%d\n%d\n%d\n%d\n", ww_size,
 			       lk_size, _HASHSIZE, _CHANNELHASHSIZE,
-			       _SERVERSIZE, poolsize);
+			       _SIDSIZE, poolsize, _UIDSIZE);
 		if (write(fd, buf, strlen(buf)) == -1)
 			sendto_flag(SCH_ERROR,
 				    "Failed (%d) to write tune file: %s.",
@@ -1255,19 +1434,18 @@ char *filename;
 /*
  * Called only from main() at startup.
  */
-void ircd_readtune(filename)
-char *filename;
+void ircd_readtune(char *filename)
 {
-	int fd, t_data[6];
+	int fd, t_data[7];
 	char buf[100];
 
 	memset(buf, 0, sizeof(buf));
 	if ((fd = open(filename, O_RDONLY)) != -1)
 	    {
 		read(fd, buf, 100);	/* no panic if this fails.. */
-		if (sscanf(buf, "%d\n%d\n%d\n%d\n%d\n%d\n", &t_data[0],
+		if (sscanf(buf, "%d\n%d\n%d\n%d\n%d\n%d\n%d\n", &t_data[0],
                            &t_data[1], &t_data[2], &t_data[3],
-                           &t_data[4], &t_data[5]) != 6)
+                           &t_data[4], &t_data[5], &t_data[6]) != 7)
 		    {
 			close(fd);
 			if (bootopt & BOOT_BADTUNE)
@@ -1288,9 +1466,12 @@ char *filename;
 		ww_size = t_data[0];
 		lk_size = t_data[1];
 		_HASHSIZE = t_data[2];
+		_HOSTNAMEHASHSIZE = t_data[2]; /* hostname has always same size
+						  as the client hash */
 		_CHANNELHASHSIZE = t_data[3];
-		_SERVERSIZE = t_data[4];
+		_SIDSIZE = t_data[4];
 		poolsize = t_data[5];
+		_UIDSIZE = t_data[6];
 
 		/*
 		** the lock array only grows if the whowas array grows,
@@ -1302,3 +1483,4 @@ char *filename;
 		close(fd);
 	    }
 }
+
